#ifndef _HASH_H
#define _HASH_H

#ifdef __cplusplus
extern "C" {
#endif

typedef unsigned long (*HashFunc)(void *);
typedef int (*HashCmp)(void *,void *);
typedef void (*HashDest)(void *);
typedef int (*HashTrav)(void *);

extern void *HashInit(HashFunc,HashCmp,unsigned long);
    /* accepts a hash function, a comparison function and a table size */
    /* the table size should be prime */
    /* returns a hash table identifier */
extern void *HashInsert(void *,void *);
    /* accepts a hash table identifier, and a user-data pointer */
    /* returns a user-data pointer if node exists, or 0 otherwise */
extern void *HashDelete(void *,void *);
    /* accepts a hash table identifier, and a user-data pointer */
    /* returns a user-data pointer if node exist(ed), or 0 otherwise */
extern void *HashSearch(void *,void *);
    /* accepts a hash table identifier and a user-data pointer */
    /* returns a user-data pointer if found, 0 otherwise */
extern int HashTraverse(void *,HashTrav);
    /* accepts a hash table identifier and a function to call with each node */
    /* if the supplied function returns false, traversal terminates, and */
    /* false is returned, otherwise, true is returned */
extern void HashDestroy(void *,HashDest);
    /* accepts a hash table identifier and a function accepting a user-data */
    /*  pointer */
    /* destroys the hash table, calling the passed function on each node */
extern void HashPrintCollisions(void *);

#ifdef __cplusplus
}
#endif

#endif
