#ifndef _WFF_H
#define _WFF_H

/* This library supports (in a simplistic way) the Wavefront Object format */

#ifdef __cplusplus
extern "C" {
#endif

#define WF_WITH_NORMALS   0x01
#define WF_WITH_TEXTURES  0x02


typedef struct _WfTriple
{
    double tr[3];
} WfTriple;

typedef struct _WfFace
{
    int flags;
    int numVerts;
    WfTriple **verts;
    WfTriple **normals;
    WfTriple **textures;
} WfFace;

typedef struct _WfObject
{
    char *fileName;
    int numVerts;
    int numNormals;
    int numTextures;
    int numFaces;
    WfTriple **verts;
    WfTriple **normals;
    WfTriple **textures;
    WfFace **faces;
} WfObject;

#ifdef __cplusplus
extern "C" {
#endif

WfObject *WfReadFile(char *);
int       WfWriteFile(char *, WfObject *);
WfObject *WfNewObject(char *); 
void      WfDeleteObject(WfObject *);

WfTriple *WfNewTriple(double, double, double);
WfTriple *WfNewTriplev(double *);
void      WfDeleteTriple(WfTriple *);

WfFace   *WfNewFace();
void      WfDeleteFace(WfFace *);
void      WfGetCenter(double[3], WfFace *);

void      WfAddVertex(WfFace *, WfTriple *, WfTriple *, WfTriple *);

#endif
