#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include <gltk.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <wff.h>
#include "glaux.h"

#define SQR(x)   (x)*(x)

const double FOV = M_PI/4;
const double GAIN_PHI = M_PI/100;
const double GAIN_THETA = M_PI/150;
const double GAIN_RADIUS = 0.05;

void myInit();
void perspective(double, double, double, double);
void myReshape(int, int);
void display();
void getCamera();
void drawAxes();
void drawObject(WfObject *);
void drawBoundingBox(double *);
void drawLine(double, double, double, double, double, double);
void getBoundingBox(WfObject *,double *);
void leftButtonDown(AUX_EVENTREC *);
void leftButtonUp(AUX_EVENTREC *);
void leftButtonLoc(AUX_EVENTREC *);
void rightButtonDown(AUX_EVENTREC *);
void rightButtonUp(AUX_EVENTREC *);
void rightButtonLoc(AUX_EVENTREC *);
void mouseMove(AUX_EVENTREC *);

static int drawWireFrame;
static WfObject *object;
static double boundingBox[7];
static int vpWidth, vpHeight;
static double camera[3];
static double radius, phi, theta;
static int lastMouseX, lastMouseY;


int main(int argc, char** argv)
{
    if (argc != 2)
    {
	fprintf(stderr, "Usage: wfview file\n");
	exit(1);
    }

    object = WfReadFile(argv[1]);
    if (!object)
    {
	fprintf(stderr, "Bad file: %s\n", argv[1]);
	exit(1);
    }
    printf(" Vertices: %d\n", object->numVerts);
    printf(" Faces: %d\n", object->numFaces);
    getBoundingBox(object, boundingBox);
    printf(" Bounding Box: (%g,%g,%g) - (%g,%g,%g)\n",
	boundingBox[0],boundingBox[1],boundingBox[2],
	boundingBox[3],boundingBox[4],boundingBox[5]);

    radius = 4;
    phi = M_PI/6;
    theta = 2*M_PI/5;

    auxInitDisplayMode(AUX_DOUBLE | AUX_RGB | AUX_DEPTH);
    auxInitPosition(0, 0, 400, 400);
    auxInitWindow(argv[0]);
    myInit();
    drawWireFrame = 0;
    auxReshapeFunc(myReshape);
    auxMouseFunc(AUX_LEFTBUTTON, AUX_MOUSEDOWN, leftButtonDown);
    auxMouseFunc(AUX_LEFTBUTTON, AUX_MOUSEUP, leftButtonUp);
    auxMouseFunc(AUX_LEFTBUTTON, AUX_MOUSELOC, leftButtonLoc);
    auxMouseFunc(AUX_RIGHTBUTTON, AUX_MOUSEDOWN, rightButtonDown);
    auxMouseFunc(AUX_RIGHTBUTTON, AUX_MOUSEUP, rightButtonUp);
    auxMouseFunc(AUX_RIGHTBUTTON, AUX_MOUSELOC, rightButtonLoc);
    auxMainLoop(display);
}


void myInit(void)
{
    GLfloat ambient[] = { 0.0, 0.0, 0.0, 1.0 };
    GLfloat diffuse[] = { 1.0, 1.0, 1.0, 1.0 };
    GLfloat specular[] = { 1.0, 1.0, 1.0, 1.0 };
    GLfloat position[] = { 0.0, 3.0, 3.0, 0.0 };
    
    GLfloat lmodel_ambient[] = { 0.2, 0.2, 0.2, 1.0 };
    GLfloat local_view[] = { 0.0 };

    GLfloat mat[4];

       /**** set lighting parameters ****/
    glLightfv(GL_LIGHT0, GL_AMBIENT, ambient);
    glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuse);
    glLightfv(GL_LIGHT0, GL_POSITION, position);
    glLightModelfv(GL_LIGHT_MODEL_AMBIENT, lmodel_ambient);
    glLightModelfv(GL_LIGHT_MODEL_LOCAL_VIEWER, local_view);
    glLightModeli( GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE );

/*    glFrontFace (GL_CW); */
    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);
    glEnable(GL_AUTO_NORMAL);
    glEnable(GL_NORMALIZE);
    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LESS);
}



void set_colour(float r, float g, float b)
{
    glDisable(GL_LIGHTING);
    glColor3f(r, g, b);
}


void set_material(float r, float g, float b)
{
  float ambient = 0.2;
  float diffuse = 0.7;
  float specular = 0.4;
  GLfloat mat[4];

    glEnable(GL_LIGHTING);

      /**** set ambient lighting parameters ****/
    mat[0] = ambient*r;
    mat[1] = ambient*g;
    mat[2] = ambient*b;
    mat[3] = 1.0;
    glMaterialfv (GL_FRONT, GL_AMBIENT, mat);

      /**** set diffuse lighting parameters ******/
    mat[0] = diffuse*r;
    mat[1] = diffuse*g;
    mat[2] = diffuse*b;
    mat[3] = 1.0;
    glMaterialfv (GL_FRONT, GL_DIFFUSE, mat);

      /**** set specular lighting parameters *****/
    mat[0] = specular*r;
    mat[1] = specular*g;
    mat[2] = specular*b;
    mat[3] = 1.0;
    glMaterialfv (GL_FRONT, GL_SPECULAR, mat);
    glMaterialf (GL_FRONT, GL_SHININESS, 0.5);
}


void display()
{
    double scalef;
    double x, y, z;
    double delta;
    double *bb = boundingBox;

    glClearColor(0,0,0,1.0);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    x = 0.5*(bb[0] + bb[3]);
    y = 0.5*(bb[1] + bb[4]);
    z = 0.5*(bb[2] + bb[5]);
    scalef = bb[3] - bb[0];
    delta = bb[4] - bb[1];
    scalef = delta>scalef ? delta : scalef;
    delta = bb[5] - bb[2];
    scalef = delta>scalef ? delta : scalef;
    glPushMatrix();
    scalef = 1./scalef;

    if (drawWireFrame)
    {
	drawAxes();
	set_colour(1, 1, 1);
	glScaled(scalef, scalef, scalef);
	glTranslated(-x, -y, -z);
	drawBoundingBox(boundingBox);
    }
    else
    {
	set_material(0.8, 0.8, 0.8);
	glScaled(scalef, scalef, scalef);
	glTranslated(-x, -y, -z);
	drawObject(object);
    }
    glPopMatrix();

    glFlush();
    tkSwapBuffers();
}


/* Why does everyone prefer the y-axis to be up? */
void getCamera()
{
    camera[2] = radius*cos(phi)*sin(theta);
    camera[0] = radius*sin(phi)*sin(theta);
    camera[1] = radius*cos(theta);
}

void myReshape(int w, int h)
{
    double aspect = h/((double)w);
    double near,far;

    vpWidth = w;
    vpHeight = h;
    glViewport(0, 0, w, h);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
     /*** this defines the field of view of the camera   ***/
     /*** Making the first 4 parameters larger will give ***/
     /*** a larger field of view, therefore making the   ***/
     /*** objects in the scene appear smaller            ***/
    if (radius-0.1 < boundingBox[6])
	near = 0.1;
    else
	near = radius - boundingBox[6];
    far = radius + boundingBox[6] + 0.1;
    perspective(FOV, aspect, near, far);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
     /*** this sets the virtual camera          ***/
     /*** gluLookAt( x,y,z,   x,y,z   x,y,z );  ***/
     /***            camera  look-at camera-up  ***/
     /***            pos'n    point   vector    ***/
    
     /*** place camera at (x=4, y=6, z=20), looking at origin,
          y-axis being up         ***/
    getCamera();
    gluLookAt(camera[0], camera[1], camera[2] ,0,0,0,0,1,0);
}


/* fov is horizontal field of view in radians */
void perspective(double fov, double aspect, double near, double far)
{
    double right = near*sin(fov/2);
    double top = right*aspect;

    glFrustum(-right, right, -top, top, near, far);
}

void drawAxes()
{
    set_colour(1, 0, 0); /* red */
    drawLine(0, 0, 0, 50, 0, 0);
    set_colour(.5, 0, 0); /* dark red */
    drawLine(0, 0, 0, -50, 0, 0);

    set_colour(0, 1, 0); /* green */
    drawLine(0, 0, 0, 0, 50, 0);
    set_colour(0, .5, 0); /* dark green */
    drawLine(0, 0, 0, 0, -50, 0);

    set_colour(0, 0, 1); /* blue */
    drawLine(0, 0, 0, 0, 0, 50);
    set_colour(0, 0, .5); /* dark blue */
    drawLine(0, 0, 0, 0, 0, -50);
}


void drawObject(WfObject *obj)
{
    WfFace *f;
    int i,j;
    int normals;

    for (i=0; i<obj->numFaces; i++)
    {
	f = obj->faces[i];
	glBegin(GL_POLYGON);
	for (j=0; j<f->numVerts; j++)
	{
	    if (f->normals[j])
		glNormal3dv(f->normals[j]->tr);
	    glVertex3dv(f->verts[j]->tr);
	}
	glEnd();
    }
}


void drawBoundingBox(double *bb)
{
    glBegin(GL_LINE_LOOP);
	glVertex3d(bb[0], bb[1], bb[2]);
	glVertex3d(bb[0], bb[4], bb[2]);
	glVertex3d(bb[3], bb[4], bb[2]);
	glVertex3d(bb[3], bb[1], bb[2]);
    glEnd();
    glBegin(GL_LINE_LOOP);
	glVertex3d(bb[0], bb[1], bb[5]);
	glVertex3d(bb[0], bb[4], bb[5]);
	glVertex3d(bb[3], bb[4], bb[5]);
	glVertex3d(bb[3], bb[1], bb[5]);
    glEnd();
    glBegin(GL_LINES);
	glVertex3d(bb[0], bb[1], bb[2]);
	glVertex3d(bb[0], bb[1], bb[5]);
	glVertex3d(bb[0], bb[4], bb[2]);
	glVertex3d(bb[0], bb[4], bb[5]);
	glVertex3d(bb[3], bb[1], bb[2]);
	glVertex3d(bb[3], bb[1], bb[5]);
	glVertex3d(bb[3], bb[4], bb[2]);
	glVertex3d(bb[3], bb[4], bb[5]);
    glEnd();
}


void drawLine(double x0,double y0,double z0,double x1,double y1,double z1)
{
    glBegin(GL_LINES);
	glVertex3d(x0, y0, z0);
	glVertex3d(x1, y1, z1);
    glEnd();
}

/* The last coordinate stores the maxiumum radius. */
void getBoundingBox(WfObject *obj,double *bb)
{
    int i;

    bb[0] = 1e30; bb[1] = 1e30; bb[2] = 1e30;
    bb[3] = -1e30; bb[4] = -1e30; bb[5] = -1e30;

    for (i=0; i<obj->numVerts; i++)
    {
	double *v = obj->verts[i]->tr;
	bb[0] = bb[0]>v[0] ? v[0] : bb[0];
	bb[1] = bb[1]>v[1] ? v[1] : bb[1];
	bb[2] = bb[2]>v[2] ? v[2] : bb[2];
	bb[3] = bb[3]<v[0] ? v[0] : bb[3];
	bb[4] = bb[4]<v[1] ? v[1] : bb[4];
	bb[5] = bb[5]<v[2] ? v[2] : bb[5];
    }
    bb[6] = 0.5*sqrt(SQR(bb[3]-bb[0]) + SQR(bb[4]-bb[1]) + SQR(bb[5]-bb[2]));
}




void leftButtonDown(AUX_EVENTREC *p)
{
    drawWireFrame = 1;
    lastMouseX = p->data[AUX_MOUSEX];
    lastMouseY = p->data[AUX_MOUSEY];
}


void leftButtonUp(AUX_EVENTREC *p)
{
    drawWireFrame = 0;
}


void leftButtonLoc(AUX_EVENTREC *p)
{
    int dx = p->data[AUX_MOUSEX] - lastMouseX;
    int dy = p->data[AUX_MOUSEY] - lastMouseY;

    lastMouseX = p->data[AUX_MOUSEX];
    lastMouseY = p->data[AUX_MOUSEY];
    phi -= dx*GAIN_PHI;
    theta += dy*GAIN_THETA;
    myReshape(vpWidth, vpHeight);
}


void rightButtonDown(AUX_EVENTREC *p)
{
    drawWireFrame = 1;
    lastMouseX = p->data[AUX_MOUSEX];
    lastMouseY = p->data[AUX_MOUSEY];
}


void rightButtonUp(AUX_EVENTREC *p)
{
    drawWireFrame = 0;
}


void rightButtonLoc(AUX_EVENTREC *p)
{
    int dx = p->data[AUX_MOUSEX] - lastMouseX;
    int dy = p->data[AUX_MOUSEY] - lastMouseY;

    lastMouseX = p->data[AUX_MOUSEX];
    lastMouseY = p->data[AUX_MOUSEY];
    radius += dy*GAIN_RADIUS;
    radius = radius<0.1 ? 0.1 : radius;
    myReshape(vpWidth, vpHeight);
}
